// gcc -Wall -Wpedantic -Wextra -std=c17 -march=haswell -O3 branchless-test.c branchless.S

#define _POSIX_C_SOURCE 199309L
#include <time.h>
#include <stdio.h>
#include <stdint.h>
#include <x86intrin.h>

#define LENGTH 64000
#define REPS 10000

// This really should be in a .h file, but I can't be bothered for a quick demo.
int64_t max(int64_t a, int64_t b);
uint64_t count_even_jump(int64_t* array, uint64_t n);
uint64_t count_even_mod(int64_t* array, uint64_t n);
uint64_t count_even_cmov(int64_t* array, uint64_t n);

void test_array_op(char* name, uint64_t(*array_operation)(int64_t*, uint64_t), int64_t* arr, uint64_t n) {
    struct timespec start, end;
    int64_t total = 0;
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start);
    for (uint64_t i = 0; i < REPS; i++) {
        total += array_operation(arr, n);
    }
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
    double elapsed = (end.tv_sec - start.tv_sec) * 1e9 + (end.tv_nsec - start.tv_nsec);
    printf("%-30s%6.0f ps/elt   Result: %ld\n", name, elapsed / (REPS * n) * 1000, total / REPS);
}

int64_t* random_array(uint64_t n) {
    int64_t* arr = (int64_t*)malloc(n * sizeof(int64_t));
    for (uint64_t i = 0; i < n; i++) {
        arr[i] = rand() % 256;
        //arr[i] = arr[i] * 2;  // make all-even array
        //arr[i] = arr[i] * 2 + 1;  // make all-odd array
    }
    return arr;
}

// A C implementation to check our assembly correctness and see how the compiler does with the problem. (spoiler: well)
uint64_t count_even_c(int64_t* array, uint64_t n) {
    uint64_t count = 0;
    for (uint64_t i = 0; i < n; i++) {
        if (array[i] % 2 == 0) {
            count += 1;
        }
    }
    return count;
}

int main(void) {
    //printf("%ld\n", max(3, 5));
    srand(time(NULL));
    int64_t* arr = random_array(LENGTH);
    test_array_op("count_even_jump", count_even_jump, arr, LENGTH);
    test_array_op("count_even_mod", count_even_mod, arr, LENGTH);
    test_array_op("count_even_cmov", count_even_cmov, arr, LENGTH);
    test_array_op("count_even_c", count_even_c, arr, LENGTH);
    free(arr);
}


